#! /usr/bin/env python

from __future__ import division
from collections import defaultdict
from operator import itemgetter
import sys
from optparse import OptionParser


class ResFileReader(object):
	"""
	Use this class to handle various file formats coming from 
	diffeernt IR engines, i.e. Terrier (answerbag) and Watson (jeopardy)
	The information we need to extract is qid, if the candidate answer is relevant or not 
	and the score IR engine assigned to this answer

	FIX: pretty up this class later on.	
	"""
	def __init__(self, mode):
		self.process_line = getattr(self, "process_line_%s" %mode)

	def process_line_jeopardy(self, line):
		tokens = line.strip().split("\t")
		qid = tokens[0]
		aid = tokens[1]
		ir_score = 1.0/(int(tokens[2]) + 1)  # we invert the score
		relevant = tokens[3]  # true or false
		return qid, aid, relevant, ir_score

	def process_line_answerbag(self, line):
		tokens = line.strip().split("\t")
		qid = tokens[0]
		aid = tokens[1]
		if qid == aid:
			relevant = "true"
		else:
			relevant = "false"
		ir_score = float(tokens[2])	
		return qid, aid, relevant, ir_score


def read_res_file(res_fname, mode="jeopardy"):
	res = open(res_fname).readlines()

	process_line = ResFileReader(mode).process_line

	ir = defaultdict(list)
	for line_res in res:
		qid, aid, relevant, ir_score = process_line(line_res)  # process the line from the res file
		ir[qid].append( (relevant, ir_score) )
	
	for qid in ir:
		ir[qid] = [rel for rel, score in sorted(ir[qid], key = itemgetter(1), reverse = True)]  # sort based on the score
	return ir


def read_res_pred_files(res_fname, pred_fname, mode="jeopardy", verbose=True):
	res = open(res_fname).readlines()
	pred = open(pred_fname).readlines()

	process_line = ResFileReader(mode).process_line

	ir, svm = defaultdict(list), defaultdict(list)
	for line_res, line_pred in zip(res, pred):
		qid, aid, relevant, ir_score = process_line(line_res)  # process the line from the res file
		pred_score = float(line_pred.strip())
		ir[qid].append( (relevant, ir_score, aid) )
		svm[qid].append( (relevant, pred_score, aid) )
	
	if verbose:
		rer_fname = res_fname + ".reranking.txt"
		file_reranking = open(rer_fname, "w")
		print "Reranking output is written to: %s" % rer_fname
		# print "%5s %8s  %2s %2s" % ("qid", "aid", "ir", "svm")
		# print "="*21

	for qid in ir:
		ir_sorted = sorted(ir[qid], key = itemgetter(1), reverse = True)
		svm_sorted = sorted(svm[qid], key = itemgetter(1), reverse = True)
		if verbose:
			ir_sorted_by_aid  = sorted([(aid, rank, rel) for rank, (rel, score, aid) in enumerate(ir_sorted)], key=itemgetter(0))
			svm_sorted_by_aid = sorted([(aid, rank, rel) for rank, (rel, score, aid) in enumerate(svm_sorted)], key=itemgetter(0))
			for (ir_aid, ir_rank, rel), (svm_aid, svm_rank, rel) in zip(ir_sorted_by_aid, svm_sorted_by_aid):
				file_reranking.write("%s %s  %2s %2s %s\n" % (qid, ir_aid, ir_rank, svm_rank, rel))
			# print "-"*21

		ir[qid] = [rel for rel, score, aid in sorted(ir[qid], key=itemgetter(1), reverse=True)]  # sort based on the score
		svm[qid] = [rel for rel, score, aid in sorted(svm[qid], key=itemgetter(1), reverse=True)]  # sort based on the score
	return ir, svm


def compute_mrr(out):
	n = len(out)
	MRR = 0.0
	for qid in out:
		relevant = out[qid]
		for i, rel in enumerate(relevant, 1):
			if rel == "true":
				MRR += 1.0/i
				break
	return MRR*100/n


def compute_precision(out, th):
	precisions = [0.0]*th
	n = 0
	for qid in out:
		relevant = out[qid]
		if all(x == "false" for x in relevant):
			continue
		for i in xrange(min(th, len(relevant))):
			if relevant[i] == "true":
				precisions[i] += 1.0
				break
		n += 1
	for i in xrange(1, th):
		precisions[i] += precisions[i-1]	

	return [p*100/n for p in precisions]


def compute_map(out, th):
	n = len(out)
	MAP = 0.0
	for qid in out:
		relevant = out[qid]
		# compute the number of relevant docs
		# get a list of precisions in the range(0,th)
		prec = [0.0]*th
		for i in xrange(min(th, len(relevant))):
			if relevant[i] == "true":
				prec[i] += 1.0
		for i in xrange(1, th):
			prec[i] += prec[i-1]	
		# print num_relevant, sum([p/i for i, p in enumerate(prec, 1)])/num_relevant, [p/i for i, p in enumerate(prec, 1)]
		# print
		MAP += sum([p/i for i, p in enumerate(prec, 1)])/th
	return MAP/n

	
def compute_accuracy(out, th):
	acc = [0.0]*th
	for qid in out:
		relevant = out[qid]
		for i in xrange(min(th, len(relevant))):
			if relevant[i] == "true":
				acc[i] += 1.0
	for i in xrange(1, th):
		acc[i] += acc[i-1]	
	
	return [a*100/(i*len(out)) for i, a in enumerate(acc, 1)]

def compute_accuracy1(out, th):
	acc = [0.0]*th
	maxrel = [0.0]*th
	for qid in out:
		relevant = out[qid]
		num_relevant = sum([1.0 for x in relevant if x == "true"])
		# print num_relevant
		for i in xrange(min(th, len(relevant))):
			if relevant[i] == "true":
				acc[i] += 1.0		
		for i in xrange(th):
			maxrel[i] += min(i+1, num_relevant)
	for i in xrange(1, th):
		acc[i] += acc[i-1]	
	return [a/numrel for a, numrel in zip(acc, maxrel)]


def compute_accuracy2(out, th):
	acc = [0.0]*th
	for qid in out:
		relevant = out[qid]
		for i in xrange(min(th, len(relevant))):
			if relevant[i] == "true":
				acc[i] += 1.0
	for i in xrange(1, th):
		acc[i] += acc[i-1]	
	return [a for i, a in enumerate(acc, 1)]


def eval_reranker(res_fname="svm.test.res", pred_fname="svm.train.pred", th=50, mode="jeopardy", verbose=False):
	try:
		ir, svm = read_res_pred_files(res_fname, pred_fname, mode, verbose)		
	except:
		print "Failed to process input files: %s %s" % (res_fname, pred_fname)
		print "Check that the input file format is correct"
		sys.exit(1)

	if mode == "jeopardy":
			# evaluate IR
		prec_se = compute_precision(ir, th)
		acc_se = compute_accuracy(ir, th)
		acc_se1 = compute_accuracy1(ir, th)
		acc_se2 = compute_accuracy2(ir, th)

		# evaluate SVM
		prec_svm = compute_precision(svm, th)
		acc_svm = compute_accuracy(svm, th)
		acc_svm1 = compute_accuracy1(svm, th)
		acc_svm2 = compute_accuracy2(svm, th)

		# map_se = compute_map(ir, th)
		# map_svm = compute_map(svm, th)

		map_se = compute_mrr(ir)
		map_svm = compute_mrr(svm)

		print "%13s %5s" %("IR", "SVM")
		print "MRRof1: %5.2f %5.2f" %(map_se, map_svm)
		# print "%16s %6s  %14s %6s  %6s %6s  %6s %6s" % ("IR", "SVM", "IR", "SVM", "IR", "SVM", "IR", "SVM")
		print "%16s %6s  %14s %6s  %14s %6s  %12s %4s" % ("IR", "SVM", "IR", "SVM", "IR", "SVM", "IR", "SVM")
		for i, (p_se, p_svm, a_se, a_svm, a_se1, a_svm1, a_se2, a_svm2) in enumerate(zip(prec_se, prec_svm, acc_se, acc_svm, acc_se1, acc_svm1, acc_se2, acc_svm2), 1):
			print "REC-1@%02d: %6.2f %6.2f  ACC@%02d: %6.2f %6.2f  AC1@%02d: %6.2f %6.2f  AC2@%02d: %4.0f %4.0f" %(i, p_se, p_svm, i, a_se, a_svm, i, a_se1, a_svm1, i, a_se2, a_svm2)
		print
		print "REC-1 - percentage of questions with at least 1 correct answer in the top @X positions (useful for tasks were questions have at most one correct answer)"
		print "ACC   - accuracy, i.e. number of correct answers retrieved at rank @X normalized by the rank and the total number of questions"
		print "AC1   - the number of correct answers at @X normalized by the number of maximum possible answers (perfect re-ranker)"
		print "AC2   - the absolute number of correct answers at @X"
	else: # answerbag
		prec_se = compute_precision(ir, th)
		prec_svm = compute_precision(svm, th)
		mrr_svm = compute_mrr(svm)
		mrr_se = compute_mrr(ir)
		print "%16s  %5s" % ("IR", "SVM")

		print "%-8s: %6.2f  %5.2f" %("MRR", mrr_se, mrr_svm)
		print
		for i, (p_se, p_svm) in enumerate(zip(prec_se, prec_svm), 1):
			print "REC-1@%02d: %6.2f %6.2f" % (i, p_se, p_svm)
		print
		print "MRR - mean reciprocal rank (http://en.wikipedia.org/wiki/Mean_reciprocal_rank)"
		print "REC-1 - percentage of questions with at least 1 correct answer in the top @X positions (useful for tasks were questions have at most one correct answer)"
	


def eval_search_engine(res_fname="svm.test.res", th=50, mode="jeopardy"):
	try:	
		ir = read_res_file(res_fname, mode)		
	except:
		print "Failed to process input file: %s" % res_fname
		print "Check that the input file format is correct"
		sys.exit(1)

	# evaluate IR
	prec = compute_precision(ir, th)
	acc = compute_accuracy(ir, th)

	if mode == "jeopardy":	# for jeopardy data we compute MAP instead of MRR since here for each query we have multiple correct answers
		MAP = compute_map(ir, th)
		print "MAP: %5.2f" % MAP
	else:  # answerbag
		MRR = compute_mrr(ir)
		print "MRR^: %5.2f%" % MRR
	
	# print stats
	print
	print "%2s: %6s  %6s" % ("#", "RECof1", "ACC")
	for i, (p, a) in enumerate(zip(prec, acc), 1):
		print "%02d: %6.2f  %6.2f" % (i, p, a)



def main():
	usage = "usage: %prog [options] arg1 arg2"
	desc = """arg1: file with the output of the baseline search engine (ex: svm.test.res) 
	arg2: predictions file from svm (ex: train.predictions)"""

	parser = OptionParser(usage=usage, description=desc)
	parser.add_option("-t", "--threshold", dest="th", default=15, type=int,
	                  help="supply a value for computing Precision up to a given threshold [default: %default]", metavar="VALUE")	 	  
	parser.add_option("-m", "--mode", dest="mode", default="answerbag", 
                  help="evalute jeopardy or answerbag [default: %default]")	 	  
	parser.add_option("-v", "--verbose", dest="verbose", default=False, action="store_true",
                  help="produce verbose output [default: %default]")	 	  
		                  

	(options, args) = parser.parse_args()

	if len(args) == 1:
		res_fname = args[0]
		eval_search_engine(res_fname, options.th, options.mode)
	elif len(args) == 2:
		res_fname = args[0]
		pred_fname = args[1]	
		eval_reranker(res_fname, pred_fname, options.th, options.mode, options.verbose)
	else:
		parser.print_help()
		sys.exit(1)
	


if __name__ == '__main__':	
	main()

