from __future__ import division
import sys
import re
import itertools
from optparse import OptionParser
import string

tokenizer = None
try:
    from nltk import PunktWordTokenizer
    tokenizer = PunktWordTokenizer()
except:
    print "No NLTK module found -- using simple word tokenizer based on spaces"



pat_chunk = re.compile(r"\[([^\[]*)\]")  # regex pattern to process the output files from chunker
# pat_qid = re.compile(r"^\d+")
transtable = string.maketrans("()", "[]")

def load_ann_data(fname):
    """
    Load annotations data file into a dict. Keys a uniq ids and values are lists of annotated sentences.

    Each annotated sentence is a dict, where keys specify the type of annotation, e.g. 'word', 'lemma', 'pos', 'chunk'
    and value is a list of corresponding annotation tokens. The length of each list must be the same.
    """
    print "Loading data from: %s..." % fname,
    data = {}
    with open(fname) as f:
        for line in f:
            qid, text = line.strip().split("\t")
            # print qid, text
            data[qid] = text
    print "done"
    return data


def read_res_file(fname):
    """
    Reads the result file of the baseline Watson search engine and returns a dict

    the file is expected to be in the following format:
    qid aid rank relevant
    """
    sys.stdout.write("Reading: %s\n" % fname)
    out = {}
    # answerkeys = {}
    with open(fname, "r") as f:
        for count, line in enumerate(f):
            tokens = line.strip().split("\t")
            qid = tokens[0]
            aid = tokens[1]
            rank = tokens[2]  # we add up 1 s.t. ranks range from 1 to 50
            relevant = "false"
            if qid == aid:
                relevant = "true"
            out.setdefault(qid, []).append((aid, rank, relevant))
            if count % 100 == 0:
                sys.stdout.write("\r%8d" % count)

    sys.stdout.write("\n")
    return out


class RerankingExampleGenerator(object):
    """
    Keeps the references to question and answers dicts
    returns string representations of examples for re-ranking
    mode specifies whether to generate examples for training or testing
    """
    def __init__(self, q_data, a_data, res, mode="train"):
        self.mode = mode
        self.q_data = q_data
        self.a_data = a_data
        self.res = res
        self.generate_examples = getattr(self, "ex_%s" % mode)
    
    def _project(self, s):
        s = s.translate(transtable)
        s = s.lower()
        if tokenizer:
            tokens = tokenizer.tokenize(s)
        else:
            tokens = s.split()
        bow_tokens = ["(%s)" % token for token in tokens if token not in string.punctuation and len(token)>=3]        
        return "(BOX %s)" % " ".join(bow_tokens)

    def _get_projections(self, question, passages):
            """
            Given a question and a list of passages returns a list of tuples where the first element is
            a projection of the passage on a question, 2nd element - projection of the question on the passage,
            3rd - rank of the passage, 4th - id of the passage.
            """
            passages_proj = []
            for aid, rank in passages:
                passage = self.a_data[aid]

                # preprocess your question and answers here
                qproj = self._project(question)  # do nothing
                aproj = self._project(passage)  # do nothing

                passages_proj.append((qproj, aproj, rank, aid))
            return passages_proj

    def _get_passages(self, qid, relevant="true"):
            """
            retrieve a list of passage ids for a given question id that are marked according to the value of argument @relevant
            and return a list of tuples with the passage id and its rank.

            Note: we skip the passages that are not present in self.a_data
            """
            return [(aid, rank) for aid, rank, rel in self.res[qid] if rel == relevant and self.a_data.get(aid, False)]

    def ex_test(self, qid):
        candidates = self.res[qid]
        q_sent = self.q_data[qid]  # take only the first sentence of the question
        for aid, cand_score, rel in candidates:  # we generate example for each candidate answer retrieved by the search engine
            candidate_answer = self.a_data.get(aid, False)
            if not candidate_answer:
                continue
            qproj = self._project(q_sent)  # do nothing
            aproj = self._project(candidate_answer)  # do nothing
            # ex = "+1\t|BT| %s |BT| %s |BT| |BT| |ET| 1:%s |BV| |EV|\n" % (qproj, aproj, cand_score)
            ex = "+1\t|BT| %s |BT| |ET| 1:%s |BV| |EV|\n" % (aproj, cand_score)
            ex_res = "%s\t%s\t%s\t%s\n" % (qid, aid, cand_score, rel)
            yield (ex, ex_res)

    def ex_train(self, qid):
        q_correct = self.q_data[qid]
        
        # get correct and incorrect passages, we ommit passages that are not in the dict with annotated data
        correct_passages = self._get_passages(qid, relevant="true")  # take only the first positive example
        incorrect_passages = self._get_passages(qid, relevant="false")

        # get a list of passage projections for the relevant and incorrect passages
        correct_passages_proj = self._get_projections(q_correct, correct_passages)
        incorrect_passages_proj = self._get_projections(q_correct, incorrect_passages)
        flip = True
        for correct, candidate in itertools.product(correct_passages_proj, incorrect_passages_proj):
            qproj_correct, aproj_correct, correct_score, pos_id = correct
            qproj_candidate, aproj_candidate, cand_score, neg_id = candidate

            if flip:
                # example = "+1\t|BT| %s |BT| %s |BT| %s |BT| %s |ET| 1:%s |BV| 1:%s |EV|\n" % (qproj_correct, aproj_correct, qproj_candidate, aproj_candidate, correct_score, cand_score)
                example = "+1\t|BT| %s |BT| %s |ET| 1:%s |BV| 1:%s |EV|\n" % (aproj_correct, aproj_candidate, correct_score, cand_score)
                example_res = "%s\t%s\t%s\t%s\t%s\n" % (qid, pos_id, neg_id, correct_score, cand_score)
            else:
                # example = "-1\t|BT| %s |BT| %s |BT| %s |BT| %s |ET| 1:%s |BV| 1:%s |EV|\n" % (qproj_candidate, aproj_candidate, qproj_correct, aproj_correct, cand_score, correct_score)
                example = "-1\t|BT| %s |BT| %s |ET| 1:%s |BV| 1:%s |EV|\n" % (aproj_candidate, aproj_correct, cand_score,  correct_score)
                example_res = "%s\t%s\t%s\t%s\t%s\n" % (qid, neg_id, pos_id, cand_score, correct_score)
            # we alternate between positive and negative pairs by flip bit
            flip = not flip
            yield (example, example_res)


def create_svm_file(qfname, afname, res_fname, options):
    q_data = load_ann_data(qfname)
    a_data = load_ann_data(afname)
    res = read_res_file(res_fname)

    mode = options.mode
    fname = options.outfname
    exfname = "%s.%s" % (fname, mode)
    exfname_res = exfname + ".res"

    # instantiate the example writer
    ex_writer = RerankingExampleGenerator(q_data, a_data, res, mode)

    count_ex = 0
    sys.stdout.write("Writing %s file to: %s\n" % (mode, exfname))
    with open(exfname, "w") as exfile:
        with open(exfname_res, "w") as exfile_res:
            for count, qid in enumerate(res, 1):
                # get a list of re-ranking pairs for a given question
                i = 0
                for i, (ex, ex_res) in enumerate(ex_writer.generate_examples(qid), 1):
                    exfile.write(ex)
                    exfile_res.write(ex_res)
                count_ex += i
                sys.stdout.write("\rQueries processed: %5d" % count)
                # break from the loop if the maximum limit is reached
                if options.limit and count >= options.limit:
                    break

    sys.stdout.write("\rQueries processed: %5d" % count)
    sys.stdout.write("\n")
    sys.stdout.write("Done! Total number of examples: %d\n" % count_ex)
    # print "Done! %d examples written to: %s" %(count, exfname)




def main():
    usage = "usage: %prog [options] questions_ann_dat answers_ann_dat res_file"
    parser = OptionParser(usage=usage)

    parser.add_option("-m", "--mode", dest="mode", default="train",
                      help="select a mode: {train, test} for generating examples file [default: %default]")
    parser.add_option("-o", "--out", dest="outfname", default="svm",
                      help="supply the output file prefix [default: %default]", metavar="FILE_PREFIX")
    parser.add_option("-l", "--limit", dest="limit", default=None, type=int,
                      help="set the maximum limit on the number of queries [default: %default]", metavar="NUM")
    # parser.add_option("-d", "--dlexical", dest="dlexical", default=1, type="int",
    #                   help="include postags of the words within the radius d around the aligned words [%default]", metavar="RADIUS")
    # parser.add_option("-c", "--dcontext", dest="dcontext", default=0, type="int",
    #                   help="include context around the aligned words within the radius d around the aligned words [%default]", metavar="RADIUS")

    (options, args) = parser.parse_args()

    if len(args) < 3:
        parser.print_usage()
        sys.exit(1)

    q_fname = args[0]
    a_fname = args[1]
    res_fname = args[2]

    create_svm_file(q_fname, a_fname, res_fname, options)

if __name__ == '__main__':
    main()
    

